#include "BST.h"

/** \brief creates a BST by creating a header cell
 *
 * \param extractKey pointer to function used for extraction of key from data
 * \param compareKeys pointer to function used to compare two keys
 * \return a pointer to the BST header
 *
 */

BST *createBST(void *(*extractKey)(void *),  int (*compareKeys)(void *, void *), void (*processData)(void *pData))
{
    BST *p = (BST *)malloc(sizeof(BST));
    if (p)
    {
        p->extractKey = extractKey;
        p->compareKeys = compareKeys;
        p->processData = processData;
        p->root = NULL;
    }
    return p;
}
/** \brief creates a BST node
 *
 * \param pData data to store at the node
 * \return pointer to newly created node
 *
 */
BSTNodeT *createNode(void *pData)
{
    BSTNodeT *p = (BSTNodeT *)malloc(sizeof(BSTNodeT));
    if (p)
    {
        p->pData = pData;
        p->parent = p->left = p->right = NULL;
    }
    return p;
}

/** \brief Insert a new data element into a BST
 *
 * \param T BST in which to insert
 * \param pData pointer to data to insert
 * \return pointer to inserted node if success or null if failure
 *
 */
BSTNodeT *insertBST(BST *T, void *pData)
{
    BSTNodeT *p, *q;
    if (!T || !pData) return NULL;
    /* build node */
    p = createNode(pData);
    if ( T->root == NULL )
    {
        /* empty tree */
        T->root = p;
        return p;
    }
    /* if we reach here then the tree is not empty ;
    look for parent node of node pointed to by p */
    q = T->root ;
    for ( ; ; )
    {
        int cmpResult = T->compareKeys(T->extractKey(pData), T->extractKey(q->pData));
        if ( cmpResult < 0 )
        {
            /* follow on left subtree */
            if ( q->left == NULL )
            {
                /* insert node here */
                q->left = p;
                p->parent = q;
                return p;
            }
            /* else */
            q = q->left ;
        }
        else if ( cmpResult > 0 )
        {
            /* follow on right subtree */
            if ( q-> right == NULL )
            {
                /* insert node here */
                q->right = p;
                p->parent = q;
                return p;
            }
            /* else */
            q = q->right ;
        }
        else
        {
            /* key already present ;
            could write a message . . . */
            free( p ) ;
            return NULL;
        }
    }
    return NULL;
}

static BSTNodeT *BSTMinimum(BSTNodeT *root)
{
    while (root->left)
        root = root->left;
    return root;
}

static BSTNodeT *BSTSuccessor(BSTNodeT *root)
{
    if (root->right)
    {
        return BSTMinimum(root->right);
    }
    BSTNodeT *y = root->parent;
    while (y != NULL && root == y->right)
    {
        y = root;
        y = y->parent;
    }
    return y;
}

// auxiliary function for performing case 0 (leaf) deletion
static BSTNodeT *deleteCase0(BSTNodeT *z)
{
    BSTNodeT *y = z->parent;
    if (y->left == z) y->left = NULL;
    else y->right = NULL;
    return z;
}

// auxiliary function for perforimng case 1 (single child) deletion
static BSTNodeT *deleteCase1(BSTNodeT *z)
{
    BSTNodeT *y = z->parent;
    BSTNodeT *succ = (z->left)? z->left: z->right;
    if (y->left == z) y->left = succ;
    else y->right = succ;
    return z;
}
/** \brief Delete a data element from a BST based ion its key
 *
 * \param T BST in which to insert
 * \param pData pointer to key of fdata element to delete
 * \return pointer to deleted node if success or null if failure
 *
 */
BSTNodeT *deleteBST(BST *T, void *pKey)
{
    if (!T || !pKey) return NULL;
    if (!T->root) return NULL;
    BSTNodeT *z = findBST(T, pKey);

    if (!z) return NULL;

    if (z->left == z->right)
    {   // leaf
        return deleteCase0(z);
    }
    else if ((z->left && !z->right) || (z->right && !z->left))
    {   // node with single child
        return deleteCase1(z);
    }
    else
    {   // swap z with its successor
        BSTNodeT *succ = BSTSuccessor(z);
        void *p = succ->pData;
        succ->pData = z->pData;
        z->pData = p;
        // perform case 0 or 1
        if (succ->left == succ->right)
            return deleteCase0(succ);
        else
            return deleteCase1(succ);
    }
}


/** \brief Find a data element from a BST based on its key
 *
 * \param T BST in which to insert
 * \param pData pointer to key of fdata element to delete
 * \return pointer to found node if success or null if failure
 *
 */

BSTNodeT *findBST(BST *T, void *pKey)
{
    if (!T || !pKey) return NULL;
    /* build node */
    if ( T->root == NULL )
    {   /* empty tree */
        return NULL;
    }
    /* if we reach here then the tree is not empty ;
    look for parent node of node pointed to by p */
    BSTNodeT *q = T->root ;
    for ( ; q != NULL; )
    {
        int cmpResult = T->compareKeys(pKey, T->extractKey(q->pData));
        if ( cmpResult < 0 )
        {
            /* follow on left subtree */
            /* else */
            q = q->left ;
        }
        else if ( cmpResult > 0 )
        {
            /* follow on right subtree */
            q = q->right ;
        }
        else
        {
            return q ;
        }
    }
    return NULL;
}

// auxiliary function for recursive preorder traversal
static void recPre(BST *T, BSTNodeT *p)
{
    if (p)
    {
        T->processData(p->pData);
        recPre(T, p->left);
        recPre(T, p->right);
    }
}

/** \brief Print BST nodes in preorder
 *
 * \param T BST to traverse
 * \return nothing
 *
 */
void preBST(BST *T)
{
    if (!T) return;
    if (!T->root) puts("Empty tree");
    else recPre(T, T->root);
    puts("");
}

// auxiliary function for recursive postorder traversal
static void recPost(BST *T, BSTNodeT *p)
{
    if (p)
    {
        recPost(T, p->left);
        recPost(T, p->right);
        T->processData(p->pData);
    }
}

/** \brief Print BST nodes in preorder
 *
 * \param T BST to traverse
 * \return nothing
 *
 */
void postBST(BST *T)
{
    if (!T) return;
    if (!T->root) puts("Empty tree");
    else recPost(T, T->root);
    puts("");
}

// auxiliary function for recuisive postorder traversal
static void recIn(BST *T, BSTNodeT *p)
{
    if (p)
    {
        recIn(T, p->left);
        T->processData(p->pData);
        recIn(T, p->right);
    }
}

/** \brief Print BST nodes in inorder
 *
 * \param T BST to traverse
 * \return nothing
 *
 */
void inBST(BST *T)
{
    if (!T) return;
    if (!T->root) puts("Empty tree");
    else recIn(T, T->root);
    puts("");
}

